" plugin/drawx.vim
"
" Drawx - Richard Bentley-Green, 26/05/2021
"
" ASCII and Unicode art drawing
"
if exists('g:Drawx_autoload_loaded')
  finish
endif
let g:Drawx_autoload_loaded = 1

" ----------------------------------
" Options

" There is also a 'second level' of 'internal' mappings defined. These must be defined with
" an additional prefix (appended to the first). This prefix must not clash with any of the
" 'trace' menu character codes below
let s:Drawx_leaderBasea = get(g:, 'Drawx_leaderBasea', '\')

" Characters to use to select various 'trace' menus
let s:Drawx_lineWeightChar = get(g:, 'Drawx_lineWeightChar', '-')
let s:Drawx_lineTypeChar = get(g:, 'Drawx_lineTypeChar', '.')
let s:Drawx_cornerChar = get(g:, 'Drawx_cornerChar', '~')
let s:Drawx_specialChar = get(g:, 'Drawx_specialChar', '+')
let s:Drawx_specialStyleChar = get(g:, 'Drawx_specialStyleChar', 's')

" 'Exit trace' key
let s:Drawx_exitTrace = get(g:, 'Drawx_exitTrace', '<esc>')

" 'Exit menu' key
let s:Drawx_exitMenu = get(g:, 'Drawx_exitMenu', '<esc>')

" Options to automatically add lines at top or bottom of buffer if required
" during tracing
let s:Drawx_autoExpandUp = get(g:, 'Drawx_autoExpandUp', 1)
let s:Drawx_autoExpandDown = get(g:, 'Drawx_autoExpandDown', 1)

" ----------------------------------
" This is a non-standard colour
highlight default link StatusMsg Normal

" Print a 'status' message to the command line
function s:PrintStatusMsg(msg)
  echohl StatusMsg | echo a:msg | echohl None
endfunction

" Print a 'warning' message to the command line
function s:PrintWarningMsg(msg)
  echohl WarningMsg | echo a:msg | echohl None
endfunction

" ------------------------------------------------------------------------------
" Various functions for converting between logical (ie, single displayed character)
" to byte-oriented column numbers, and for retrieving and setting the cursor position.
" These functions are needed to facilitate correct rendering of multi-byte
" Unicode characters
"
" IMPORTANT: All line and column numbers are 1-based (even the 'byte' version). Zero
"            may be used in some cases to indicate an error or some special state

" ----------------------------------
" Convert a byte-oriented column number to a logical column number
"
" Returns logical column number. Zero is returned on an error
"
function s:bColTolCol(line, bcol)
  return virtcol([a:line, a:bcol])
endfunction

" Convert a 1-based logical column number to a byte-oriented column number
" If the specified column is beyond the end of the line then the result will
" indicate the end of the line that IS present
"
" Returns byte-oriented column number. Zero is returned on an error
"
function s:lColTobCol(line, lcol)
  return strlen(strcharpart(getline(a:line), 0, a:lcol - 1)) + 1
endfunction

" Get cursor position, returning logical column number
"
" posn -  list of 2 elements [line, lcol]
"
function s:getCursorPosn()
  let l:posn = getpos('.')
  let l:posn[2] = virtcol([l:posn[1], l:posn[2] + l:posn[3]])
  return l:posn[1:2]
endfunction

" Return indication of whether the specified position is in virtual space or not
"
" line - Line and logical column to test
" lcol
"
" Returns
" Returns 1 if position is in virtual space, else returns 0
"
function s:isPosnVirtual(line, lcol)
  let l:virt = 0

  " Calculate logical length of line
  let l:line = getline(a:line)
  let l:len = strlen(strcharpart(l:line, 0, a:lcol - 1))
  let l:llen = strchars(strcharpart(l:line, 0, a:lcol))
  if (a:lcol > l:llen)
    let l:virt = 1
  endif

  return l:virt
endfunction

" Set cursor position, using logical column number. If the line is not long
" enough to accommodate the specified column then it is appended with spaces
" as required (so the cursor is left at the last space in the line).
" As a:lcol must be > 0, if the line is empty then 1 space shall be appended
" to the line
"
" line - Line and logical column to position cursor
" lcol
"
" Returns
" If line has been extended to accomodate new position, returns 1, else returns 0
"
function s:setCursorPosnExt(line, lcol)
  let l:extended = 0

  " Calculate logical length of line
  let l:line = getline(a:line)
  let l:len = strlen(strcharpart(l:line, 0, a:lcol - 1))
  let l:llen = strchars(strcharpart(l:line, 0, a:lcol))
  if (a:lcol > l:llen)
    " Line not long enough - pad it. This will leave the cursor in the correct position
    exe 'normal! '.a:line.'G'.(a:lcol - l:llen)."A \<esc>"
    let l:extended = 1
  else
    " Line is long enough - just position cursor
    "let l:posn = [0, a:line, l:len + 1, 0]
    "call setpos('.', l:posn)
    call cursor(a:line, l:len + 1)
  endif

  return l:extended
endfunction

" Return the character at the cursor position. If the cursor is in virtual space
" then it shall return the last character on the line. If the line is blank then
" it shall return ''
"
function s:getCharAtCursor()
  " Apparently faster that the 'matchstr' versions
  return strcharpart(strpart(getline('.'), col('.') - 1), 0, 1)
endfunction

" ------------------------------------------------------------------------------
" Return visual mode ordinates
"
" rat - If '1' then the returned list elements are ordered so that line1 <=
"       line2 and col1 <= col2 (rationalised).
"       If '0' then the returned list elements are ordered so that the
"       line1/col1 represents the cursor position at the start of the selection
"       and line2/col2 represents the current cursor position at the end of the
"       selection (ie, the current cursor position)
"
" Returns
" A list in the order [line1, col1, line2, col2]. An empty list is returned on
" an error. The 'col' values are logical, NOT byte-oriented
"
function s:getVOrdinates(rat)
  " Get start and end line numbers of the visual selection
  let [l:l1, l:c1a, l:c1b] = getpos("'<")[1:3]
  let [l:l2, l:c2a, l:c2b] = getpos("'>")[1:3]
  let l:posn = []

  " The act of handling the key mapping will have caused visual mode to drop-out. Return to it
  exe 'normal! gv'

  if (mode() != '')
    " We are in a visual mode other than block - not usable
    call s:PrintWarningMsg('Cannot use non-block visual mode to draw')
  else
    " Get the logical column numbers of the start and end of the visual selection
    let l:c1 = virtcol("'<")
    let l:c2 = virtcol("'>")

    if a:rat == 1
      " Rationalise the returned list
      let l:posn = [l:l1, (l:c1 >= l:c2) ? l:c2 : l:c1, l:l2, (l:c1 >= l:c2) ? l:c1 : l:c2]
    else
      " Return list in an order showing the actual selection start and end points.
      " l1 will always be <= l2, so we need to check the current cursor position
      " to work out the order to return the list (to set cursor position at the end)
      let l:cursor = getpos('.')
      if ((l:cursor[1] == l:l1) && (l:cursor[2] == (l:c1a + l:c1b)))
        " Cursor is on line l1 (actually, position '<)
        let l:posn = [l:l2, l:c2, l:l1, l:c1]
      else
        " Cursor is on line l2 (actually position '>)
        let l:posn = [l:l1, l:c1, l:l2, l:c2]
      endif
    endif

    " Drop out of visual selection mode again
    exe "normal! \<esc>"
  endif

  return l:posn
endfunction

" ------------------------------------------------------------------------------
" Styles

let s:lineWeightChars = ['-', '─', '━', '═', '•', ' ']
let s:lineTypeChars = ['─', '╌',  '┄',  '┈']
let s:cornerTypeChars = ['┌', '╭']

" ASCII
let s:styleASCII = ['-', '|', '+', '+', '+', '+']

" Thin
let s:styleThin = ['─', '│', '┌', '┐', '┘', '└']
let s:styleThinDashed = ['╌', '╎', '┄', '┆', '┈', '┊']
let s:styleThinRound = ['╭', '╮', '╯', '╰']

" Thick
let s:styleThick = ['━', '┃', '┏', '┓', '┛', '┗']
let s:styleThickDashed = ['╍', '╏', '┅', '┇', '┉', '┋']

" Double-line
let s:styleDbl = ['═', '║', '╔', '╗', '╝', '╚']

" Dotted
let s:styleDot = ['•', '•', '•', '•', '•', '•']

" Blank/Erase
let s:styleBlank = [' ', ' ', ' ', ' ', ' ', ' ']

" Tables for determining how a 't', 'T', 'x' or 'X' is rendered
"
" Example of how this works;-
" Say the operation is a 't', the direction is down, the current style is thin
" and the current character at the current position is '━' (ie, a thick line)
"
" 1/ If the current character is one of the dashed lines, then convert it to a
"    non-dashed line, using the s:styleThinDashed and s:styleThickDashed tables
"    above (in this example, the current character is not dashed)
" 2/ As the direction is up then we search s:joinH (because we're
"    (potentially) joining to a horizontal line) for the current character.
" 3/ We find the current character at index 2 of s:joinH
" 4/ Because the current style is thin, we select s:thinH. Because the direction
"    is up and the operation is a 't', we select s:thinH[0][2 + 1] which is our
"    final character to place
" 5/ If the initial search of joinH (at (3)) yielded no result (say if the current
"    character were ' '), then an index of -1 would be generated, and so we would
"    use the default 'up T' for the thin style which is s:thinH[0][0] (0 = -1 + 1)

" Lookup tables used when drection is left or right
let s:joinV = [
  \ ['|', '│', '┃', '║', '•'],
  \ ['+', '├', '┠', '╟'], ['', '┝', '┣', '╠'], ['', '╞'],
  \ ['',  '┤', '┨', '╢'], ['', '┥', '┫', '╣'], ['', '╡'],
  \ ['',  '┼', '╂', '╫'], ['', '┿', '╋', '╬'], ['', '╪']]

"                Lt T                            Rt T                            X
let s:ASCIIV =  [['+', '+', '├', '┠', '╟', '+'], ['+', '+', '┤', '┨', '╢', '+'], ['+', '+', '┼', '╂', '╫', '+']]
let s:thinV =   [['├', '├', '├', '┠', '╟', '├'], ['┤', '┤', '┤', '┨', '╢', '┤'], ['┼', '┼', '┼', '╂', '╫', '┼']]
let s:thickV =  [['┣', '┝', '┝', '┣', '╠', '┣'], ['┫', '┥', '┥', '┫', '╣', '┫'], ['╋', '┿', '┿', '╋', '╬', '╋']]
let s:DblV =    [['╠', '╞', '╞', '╠', '╠', '╠'], ['╣', '╡', '╡', '╣', '╣', '╣'], ['╬', '╪', '╪', '╬', '╬', '╬']]
let s:DotV =    [['•', '+', '├', '┣', '╠', '•'], ['•', '+', '┤', '┫', '╣', '•'], ['•', '+', '┼', '╋', '╬', '•']]

let s:BlankVH = [[' ', ' ', ' ', ' ', ' ', ' '], [' ', ' ', ' ', ' ', ' ', ' '], [' ', ' ', ' ', ' ', ' ', ' ']]

let s:joinVindex = [s:ASCIIV, s:thinV, s:thickV, s:DblV, s:DotV, s:BlankVH]

" Lookup tables used when drection is up or down
let s:joinH = [
  \ ['-', '─', '━', '═', '•'],
  \ ['+', '┬', '┯', '╤'], ['', '┰', '┳', '╦'], ['', '╥'],
  \ ['',  '┴', '┷', '╧'], ['', '┸', '┻', '╩'], ['', '╨'],
  \ ['',  '┼', '┿', '╪'], ['', '╂', '╋', '╬'], ['', '╫']]

"                Up T                            Dn T                            X
let s:ASCIIH = [['+', '+', '┬', '┯', '╤', '+'], ['+', '+', '┴', '┷', '╧', '+'], ['+', '+', '┼', '┿', '╪', '+']]
let s:thinH =  [['┬', '┬', '┬', '┯', '╤', '┬'], ['┴', '┴', '┴', '┷', '╧', '┴'], ['┼', '┼', '┼', '┿', '╪', '┼']]
let s:thickH = [['┳', '┰', '┰', '┳', '╦', '┳'], ['┻', '┸', '┸', '┻', '╩', '┻'], ['╋', '╂', '╂', '╋', '╬', '╋']]
let s:DblH =   [['╦', '╦', '╥', '╦', '╦', '╦'], ['╩', '╩', '╨', '╩', '╩', '╩'], ['╬', '╫', '╫', '╬', '╬', '╬']]
let s:DotH =   [['•', '+', '┬', '┳', '╦', '•'], ['•', '+', '┴', '┻', '╩', '•'], ['•', '+', '┼', '╋', '╬', '•']]

let s:joinHindex = [s:ASCIIH, s:thinH, s:thickH, s:DblH, s:DotH, s:BlankVH]

" Area fills
let s:styleFills = [' ']

let s:specialChars = ['■', '□', '○', '◌', '◍', '◯', '←', '↑', '→', '↓', '⇄', '⇆', '⇅', '⋯', '⋮', '∶', '∷']

" ------------------------------------------------------------------------------
" Translate character lookup tables

" Flip characters top-to-bottom
let s:transV = [
  \ ['┌', '┐', '╭', '╮', '┏', '┓', '╔', '╗',    '┬', '┯', '╤', '┰', '┳', '╥', '╦',   '↑'],
  \ ['└', '┘', '╰', '╯', '┗', '┛', '╚', '╝',    '┴', '┷', '╧', '┸', '┻', '╨', '╩',   '↓']]
    " Thin               Thick     Double-line  Thin-T         Thick-T   Double-l-T  Special

" Flip characters left-to-right
let s:transH = [
  \ ['┌', '└', '╭', '╰', '┏', '┗', '╔', '╚',    '├', '┠', '╟', '┝', '┣', '╠', '╞',   '←'],
  \ ['┐', '┘', '╮', '╯', '┓', '┛', '╗', '╝',    '┤', '┨', '╢', '┥', '┫', '╣', '╡',   '→']]
    " Thin               Thick     Double-line  Thin-T         Thick-T   Double-l-T  Special

" ------------------------------------------------------------------------------
" Translate a string; flipping selected characters to other characters
"
" l:text - text to translate
" tbl - Translation table to lookup in
"
function s:translate(text, tbl)
  let l:text = ''
  if a:text != ''
    let l:text = split(a:text, '\zs')

    let l:x = 0
    while l:x < len(l:text)
      let l:idx = index(a:tbl[0], l:text[l:x])
      if l:idx >= 0
        let l:text[l:x] = a:tbl[1][l:idx]
      else
        let l:idx = index(a:tbl[1], l:text[l:x])
        if l:idx >= 0
          let l:text[l:x] = a:tbl[0][l:idx]
        endif
      endif

      let l:x += 1
    endwhile

    let l:text = join(l:text, '')
  endif

  return l:text
endfunction

" ------------------------------------------------------------------------------
" The currently selected style
let s:csAll = s:styleThin
let s:csLines = s:csAll[0:1]
let s:csCorners = s:csAll[2:5]
let s:csFill = s:styleFills[0]

" line weight - 0 = ASCII, 1 = thin, 2 = thick, 3 = double, 4 = dotted, 5 = blank
let s:csWeight = get(g:, 'Drawx_defaultLineWeight', 1)

" Indication that the selected style can form a continuous line. The
" alternative, is a dotted (or other character) line
let s:csCont = 1

" line type   1 = solid, 2 = 2-dash, 3 = 3-dash, 4 = 4-dash
" corner type 1 = square, 2 = round
"
" NOTE: In both cases, these are what the user requested but are NOT necessarily
"       what is in force (availability depends on the s:csWeight selected)
let s:line = get(g:, 'Drawx_defaultLineType', 1)
let s:corner = get(g:, 'Drawx_defaultCornerType', 1)

" The current style
let s:cs = [s:csAll, s:csLines, s:csCorners, s:csFill, s:csWeight, s:csCont]

" 'Erase mode' setting (0 = off, 1 = on)
let s:eraseMode = 0

" ------------------------------------------------------------------------------
" Set the style
"
" weight - line weight - 0 = ASCII, 1 = thin, 2 = thick, 3 = double, 4 = dotted, 5 = blank
" line - line type - 1 = solid, 2 = 2-dash, 3 = 3-dash, 4 = 4-dash
" corner - corner type - 1 = square, 2 = round
"
function s:setStyle(weight, line, corner)
  " Line weight
  if a:weight >= 0
    let s:csCont = 1
    if a:weight == 0
      let s:csAll = s:styleASCII
    elseif a:weight == 1
      let s:csAll = s:styleThin
    elseif a:weight == 2
      let s:csAll = s:styleThick
    elseif a:weight == 3
      let s:csAll = s:styleDbl
    elseif a:weight == 4
      let s:csAll = s:styleDot
      let s:csCont = 0
    else
      let s:csAll = s:styleBlank
    endif

    let s:csWeight = a:weight
  endif

  " Line type
  if a:line > 0
    let s:line = a:line
  endif

  if s:line > 1
    let l:line = (s:line - 2) * 2
    if s:csWeight == 1
      let s:csLines = s:styleThinDashed[l:line:l:line + 1]
    elseif s:csWeight == 2
      let s:csLines = s:styleThickDashed[l:line:l:line + 1]
    else
      let s:csLines = s:csAll[0:1]
    endif
  else
    let s:csLines = s:csAll[0:1]
  endif

  " Corner
  if a:corner > 0
    let s:corner = a:corner
  endif

  " Round corners are only supported for 'thin' style
  if (s:corner > 1) && (s:csWeight == 1)
    let s:csCorners = s:styleThinRound
  else
    let s:csCorners = s:csAll[2:5]
  endif

  " Only establish new style if the weight is defined or if the current style is
  " NOT constructed from a character (via a call to s:setStyleFromChar())
  if (a:weight >= 0) || (s:cs[4] >= 0)
    " Establish style
    let s:cs = [s:csAll, s:csLines, s:csCorners, s:csFill, s:csWeight, s:csCont]

    call s:PrintStatusMsg("Active style - Line weight:'".s:lineWeightChars[s:csWeight]."' Line type:'".s:lineTypeChars[s:line - 1]."' Corner type:'".s:cornerTypeChars[s:corner - 1])
  else
    call s:PrintStatusMsg('Active style - Still set based on "'.s:cs[0][0].'"')
  endif
endfunction

" ------------------------------------------------------------------------------
"  Draw a box
"
" posn - The area to draw into
" style - The style to draw with
" fill - 0 = do not fill. 1 = fill with space characters
"
function s:drawBox(posn, style, fill)
  " Note: These width and height are actually 1 smaller than actual
  let l:height = a:posn[2] - a:posn[0]
  let l:width = a:posn[3] - a:posn[1]

  " Starting point
  let l:line = a:posn[0]
  let l:col = a:posn[1]

  " Draw top
  call s:setCursorPosnExt(l:line, l:col)
  exe 'normal! R'.a:style[2][0]." \<esc>"
  if l:width > 1
    exe 'normal! '.(l:width).'R'.a:style[1][0]."\<esc>R \<esc>"
  endif
  exe 'normal! R'.a:style[2][1]."\<esc>"

  " Draw middle
  let l:h = l:height
  if a:fill
    while l:h > 1
      let l:line += 1
      call s:setCursorPosnExt(l:line, l:col)
      exe 'normal! R'.a:style[1][1]." \<esc>R \<esc>"
      if l:width > 1
        exe 'normal! '.(l:width).'R'.a:style[3]."\<esc>R \<esc>"
      endif
      exe 'normal! R'.a:style[1][1]."\<esc>"
      let l:h -= 1
    endwhile
  else
    while l:h > 1
      let l:line += 1
      call s:setCursorPosnExt(l:line, l:col)
      exe 'normal! R'.a:style[1][1]."\<esc>"
      call s:setCursorPosnExt(l:line, l:col + l:width)
      exe 'normal! R'.a:style[1][1]."\<esc>"
      let l:h -= 1
    endwhile
  endif

  " Draw bottom
  let l:line += 1
  call s:setCursorPosnExt(l:line, l:col)
  exe 'normal! R'.a:style[2][3]." \<esc>"
  if l:width > 1
    exe 'normal! '.(l:width).'R'.a:style[1][0]."\<esc>R \<esc>"
  endif
  exe 'normal! R'.a:style[2][2]."\<esc>"
endfunction

" ------------------------------------------------------------------------------
" Draw a box (visual mode)
"
" Invoked in visual selection mode - the selection co-ordinates indicate the
" size of the box
"
function drawx#vDrawBox(fill) range
  let l:posn = s:getVOrdinates(1)
  if !empty(l:posn) && (l:posn[0] != l:posn[2]) && (l:posn[1] != l:posn[3])
    call s:drawBox(l:posn, s:cs, a:fill)
  else
    call s:PrintWarningMsg('Visual selection too small to draw into')
  endif
endfunction

" ------------------------------------------------------------------------------
" Replace character at specified cursor position
"
" line
" col
" char
"
function s:setChar(line, col, char)
  call s:setCursorPosnExt(a:line, a:col)
  exe 'normal! r'.a:char."\<esc>"
endfunction

" ------------------------------------------------------------------------------
" Draw an ellipse
"
" This function is rather complicated. The reason for the complexity is that
" drawing an ellipse on a low resolution text display and making it look good
" in all cases turns out to be quite awkward as there are a few edge cases
" that would otherwise spoil the final image. So although this is complicated,
" it produces good results. Some of the logic could possibly be simplified but
" if it's not broken...
"
" posn - The area to draw into
" style - The style to draw with
" vinflate - 0 or 1. The '0' version is mathematically more accurate but tends
"            to be more of a 'spiky' ellipse. The '1' ("inflated") version makes
"            a 'rounder' ellipse
" fill - 0 = do not fill. 1 = fill with space characters
"
function s:drawEllipse(posn, style, vinflate, fill)
  " Note: These width and height are actually 1 smaller than actual
  let l:height = (a:posn[2] - a:posn[0]) + (a:vinflate * 2)
  let l:width = a:posn[3] - a:posn[1]
  let l:halfHeight = l:height / 2
  let l:halfWidth  = l:width / 2

  " Centre point
  let l:cline = a:posn[0] + l:halfHeight - a:vinflate
  let l:ccol = a:posn[1] + l:halfWidth

  let l:hh = l:halfHeight * l:halfHeight
  let l:ww = l:halfWidth * l:halfWidth
  let l:hhww = l:hh * l:ww
  let l:x0 = l:halfWidth
  let l:dx = 0

  " Indication of whether or not line is continuous - effects shape and fill
  let l:style_is_linetype = a:style[5]

  " If the width is even then we need to insert a column down the centre of
  " the drawn ellipse, and if the height is even, we need to insert a row
  " (bearing in mind 'l:width' and 'l:height' are 1 smaller than actual)
  let l:hPad = and(l:width, 1)
  let l:vPad = and(l:height, 1)

  " Calculate and note the length of each line, working out from the centre
  let l:lines = []

  " Add centre line (+1 to include centre column in length)
  call add(l:lines, l:halfWidth + 1)

  let l:y = 1

  " Calculate initial slope of circumference
  let l:x1 = l:x0 - (l:dx - 1)
  while (l:x1 > 0) && (((l:x1 * l:x1 * l:hh) + (l:y * l:y * l:ww)) > l:hhww)
    let l:x1 -= 1
  endwhile

  let l:prevx1 = l:halfWidth
  let l:nextx1 = 0

  while l:y <= (l:halfHeight - a:vinflate)
    let l:dx = l:x0 - l:x1
    let l:x0 = l:x1

    " Calculate slope of circumference for NEXT line(s). This look-ahead is used
    " below to conditionally modify the shape of the drawn object
    let l:nextx1 = (l:y == l:halfHeight) ? 0 : l:x0 - (l:dx - 1)
    while (l:nextx1 > 0) && (((l:nextx1 * l:nextx1 * l:hh) + ((l:y + 1) * (l:y + 1) * l:ww)) > l:hhww)
      let l:nextx1 -= 1
    endwhile

    " +1 to include the centre column in the length
    let l:length = l:x0 + 1

    " Override length for special case of first line (next to centre line) if a short but wide ellipse
    " with only a single centre line. It's not possible to draw (correctly) the left and right edges
    " of the ellipse if we don't do this in this case
    if !l:vPad && (l:y == 1)
      let l:length = l:halfWidth
    endif

    " In certain cases, increase the line length by 2. The calculated length
    " isn't wrong but doesn't look ideal when rendered in (low resolution) text
    " (the result can look more hexagonal than round).
    if (l:prevx1 && l:nextx1 && (l:nextx1 == l:x1) && (l:x1 < l:prevx1)) || ((l:prevx1 == 1) && !l:x1 && !l:nextx1 && ((l:y < (l:halfHeight - a:vinflate)) || (l:style_is_linetype && !l:hPad))) || (!l:hPad && (l:length == 1) && l:style_is_linetype) || ((l:y == 1) && l:style_is_linetype && !l:vPad)
      let l:x1 += 1
      let l:length += 1
    endif

    " Note the line length
    call add(l:lines, l:length)

    let l:prevx1 = l:x1
    let l:x1 = l:nextx1

    let l:y += 1
  endwhile

  " Add a dummy null line to make the following easier
  call add(l:lines, 0)

  " Rationalise the height to avoid having to do this throughout the following
  let l:halfHeight += (1 - a:vinflate)

  " Draw each line, working from the centre outwards
  let l:y = 0
  while l:y < l:halfHeight
    let l:x = 0
    let l:len = l:lines[l:y]
    let l:nextlen = (l:y < l:halfHeight) ? l:lines[l:y + 1] : 0

    " Never any fill for top/bottom lines
    let fill = (l:y < (l:halfHeight - 1))
    while l:x < l:len
      if l:fill
        " FILL
        if a:fill
          " If the next line is only 2 characters wide then there cannot be any fill on this line
          if (!l:style_is_linetype || (l:nextlen > 1))
            call s:setChar(l:cline - l:y, l:ccol - l:x, ' ')
            call s:setChar(l:cline - l:y, l:ccol + l:hPad  + l:x, ' ')

            if (l:vPad || l:y)
              call s:setChar(l:cline + l:vPad  + l:y, l:ccol - l:x, ' ')
              call s:setChar(l:cline + l:vPad  + l:y, l:ccol + l:hPad + l:x, ' ')
            endif

            if (l:style_is_linetype && (l:len > l:nextlen) && (l:x == (l:nextlen - 1))) || (l:style_is_linetype && (l:len >= l:nextlen) && (l:x == (l:nextlen - 2))) || (!l:style_is_linetype && (((l:len > l:nextlen) && (l:x == (l:nextlen - 1))) || ((l:len == l:nextlen) && (l:x == (l:nextlen - 2)))))
              " End of fill for this line
              let l:fill = 0
            endif

            let l:x += 1
          else
            " End of fill for this line
            let l:fill = 0
          endif
        else
          " No fill at all to draw - skip to circumference drawing position
          if l:style_is_linetype
            let l:x = (l:len > l:nextlen) ? l:nextlen - 1 : l:len - 1
          else
            let l:x = (l:len > l:nextlen) ? l:nextlen : l:len - 1
          endif

          let l:fill = 0
        endif
      else
        " CIRCUMFERENCE
        let l:c0 = 1
        let l:c1 = [0, 0, 0, 0]
        if l:x == (l:len - 1)
          " Outer-edge character
          let l:c1 = [1, 1, 1, 1]
          if (l:y || (!l:y && l:vPad)) && (l:len > l:nextlen)
            " Draw corners to 'cap off' the end of the (2) centre line
            let l:c0 = 2
            let l:c1[0] = 0
            let l:c1[1] = 3
            let l:c1[2] = 1
            let l:c1[3] = 2
          endif
        elseif l:style_is_linetype && (l:len > l:nextlen) && (l:x == (l:nextlen - 1))
          " Join to next line
          let l:c0 = 2
          let l:c1[0] = 2
          let l:c1[1] = 1
          let l:c1[2] = 3
          let l:c1[3] = 0
        endif

        call s:setChar(l:cline - l:y, l:ccol - l:x, a:style[l:c0][l:c1[0]])
        call s:setChar(l:cline + l:vPad  + l:y, l:ccol - l:x, a:style[l:c0][l:c1[1]])
        call s:setChar(l:cline - l:y, l:ccol + l:hPad  + l:x, a:style[l:c0][l:c1[2]])
        call s:setChar(l:cline + l:vPad  + l:y, l:ccol + l:hPad + l:x, a:style[l:c0][l:c1[3]])

        let l:x += 1
      endif
    endwhile
    let l:y += 1
  endwhile
endfunction

" ------------------------------------------------------------------------------
" Draw an ellipse (visual mode)
"
" Invoked in visual selection mode - the selection co-ordinates indicate the
" size of the ellipse
"
function drawx#vDrawEllipse(vinflate, fill) range
  let l:posn = s:getVOrdinates(1)
  if !empty(l:posn) && ((l:posn[2] - l:posn[0]) >= 2) && ((l:posn[3] - l:posn[1]) >= 2)
    call s:drawEllipse(l:posn, s:cs, a:vinflate, a:fill)
  else
    call s:PrintWarningMsg('Visual selection too small to draw into')
  endif
endfunction

" ------------------------------------------------------------------------------
" Clear an area (overwrite with space characters)
"
" posn - Position and area to clear
"
function s:Clear(posn)
  let l:line = a:posn[0]
  let l:col = a:posn[1]
  let l:width = (a:posn[3] - l:col) + 1
  while l:line <= a:posn[2]
    call s:setCursorPosnExt(l:line, l:col)
    exe 'normal! '.(l:width)."R \<esc>"
    let l:line += 1
  endwhile
endfunction

" ------------------------------------------------------------------------------
" Yank (and optionally clear) (visual mode)
"
" clear - 0 = yank only, 1 = yank and clear
"
function drawx#vYank(clear) range
  " Make a note of v:register because it will be lost as a result of callling getVOrdinates()
  let l:register = v:register
  let l:posn = s:getVOrdinates(1)
  if !empty(l:posn)
    " Return to visual selection mode, yank and then drop out of visual
    " selection mode again
    exe 'normal! gv"'.l:register."y\<esc>"

    if a:clear
      call s:Clear(l:posn)
    endif

    call s:PrintStatusMsg("Yanked area W:".((l:posn[3] - l:posn[1]) + 1).', H:'.((l:posn[2] - l:posn[0]) + 1))
  endif
endfunction

" ------------------------------------------------------------------------------
" Delete/clear (visual mode)
"
function drawx#vClear() range
  let l:posn = s:getVOrdinates(1)
  if !empty(l:posn)
    call s:Clear(l:posn)
    call s:PrintStatusMsg("Cleared area W:".((l:posn[3] - l:posn[1]) + 1).', H:'.((l:posn[2] - l:posn[0]) + 1))
  endif
endfunction

" ------------------------------------------------------------------------------
" Paste
"
" register - The register to paste from
" posn - Position and area to paste into
" part - Indicates how the clip is to be pasted if its size does not exactly
"        match that of the postions's area. If the area is bigger than the clip
"        then the following indicates where in the position's area the cliip
"        should be pasted. If the position's area is smaller than the clip then
"        the following indicates which part of the clip should be pasted. In all
"        cases, the width and height are assesed seperately;-
"        1 = top-left
"        2 = top-right
"        3 = bottom-right
"        4 = bottom-left
" cutout - 1 = omit leading and trailing whitespace from the pasted clip, 0 =
"          past all the clip (generally resulting in a rectangular paste area)
" ud - 1 = Paste clip upside-down, 0 otherwise
" bf - 1 = Paste clip back-to-front, 0 otherwise
"
function s:Paste(register, posn, part, cutout, ud, bf)
  " Get 'paste' data from selection register
  let l:yankBuff = []
  let l:yankHeight = 0
  let l:yankWidth = 0

  if getregtype(a:register) =~# "^\<C-V>[1-9]"
    " Block selection
    let l:yankBuff = split(getreg(a:register), '\n')
    let l:yankHeight = len(l:yankBuff)
    if l:yankHeight
      let l:yankWidth = str2nr(strpart(getregtype(a:register), 1))
    endif
  elseif getregtype(a:register) ==# 'v'
    " Character-wise selection
    let l:yankBuff = split(getreg(a:register), '\n')
    let l:yankHeight = len(l:yankBuff)
    if l:yankHeight == 1
      " Selection is on a single line
      let l:yankWidth = strchars(l:yankBuff[0])
    else
      " Can't use this selection
      let l:yankHeight = 0
      call s:PrintWarningMsg("Cannot paste non-single line 'char' selection")
    endif
  else
    call s:PrintWarningMsg("Cannot paste non-block/non-single line 'char' selection")
  endif

  if l:yankHeight
    " Size of area to paste into
    let l:pasteHeight = (a:posn[2] - a:posn[0]) + 1
    let l:pasteWidth = (a:posn[3] - a:posn[1]) + 1

    " Determine start and end points in clip that shall be pasted
    let l:yl = 0
    let l:yr = l:yankWidth - 1
    let l:yt = 0
    let l:yb = l:yankHeight - 1

    if l:yankHeight > l:pasteHeight
      if a:part < 3
        let l:yb = l:pasteHeight - 1
      else
        let l:yt = l:yankHeight - l:pasteHeight
      endif
    endif

    if l:yankWidth > l:pasteWidth
      if (a:part == 1) || (a:part == 4)
        let l:yr = l:pasteWidth - 1
      else
        let l:yl = l:yankWidth - l:pasteWidth
      endif
    endif

    " Determine start and end points in buffer that shall be pasted into
    let l:line = a:posn[0]
    let l:col = a:posn[1]

    if (l:yankHeight < l:pasteHeight) && (a:part > 2)
      let l:line += (l:pasteHeight - l:yankHeight)
    endif

    if (l:yankWidth < l:pasteWidth) && ((a:part == 2) || (a:part == 3))
      let l:col += (l:pasteWidth - l:yankWidth)
    endif

    " Paste
    let l:py = (a:ud) ? l:yb : l:yt
    let l:pend = (a:ud) ? l:yt - 1 : l:yb + 1
    let l:pinc = (a:ud) ? -1 : 1

    while l:py != l:pend
      let l:text = strcharpart(l:yankBuff[l:py], l:yl, (l:yr - l:yl) + 1)
      if a:ud
        " Translate string for upside-down pasting
        let l:text = s:translate(l:text, s:transV)
      endif

      if a:bf
        " Reverse string for back-to-front pasting and then translate it
        let l:text = join(reverse(split(l:text, '\zs')), '')
        let l:text = s:translate(l:text, s:transH)
      endif

      if a:cutout
        " Do not paste leading and trailing whitespace in the line
        let l:off = match(l:text, '[^ ]')
        if l:off >= 0
          " Line is not completely blank (all spaces)
          let l:len = match(l:text, '  *$')
          if l:len < 0
            let l:len = strlen(l:text) - l:off
          else
            let l:len -= l:off
          endif

          " Note that l:col is a logical column number and l:off is a byte offset.
          " This is ok because l:off is effectively the number of space characters
          " at the start of the clip, and as a space character is a single byte,
          " the calculation still works - ie, converting l:off to a number of
          " characters rather than bytes (to make it 'logical') would yield the
          " same value
          call s:setCursorPosnExt(l:line, l:col + l:off)
          exe 'normal! R'.strpart(l:text, l:off, l:len)."\<esc>"
        endif
      else
        " Paste the whole (rectangular) snip
        call s:setCursorPosnExt(l:line, l:col)
        exe 'normal! R'.l:text."\<esc>"
      endif

      let l:py += l:pinc
      let l:line += 1
    endwhile

    call s:PrintStatusMsg('Pasted area W:'.((l:yr - l:yl) + 1).', H:'.((l:yb - l:yt) + 1))
  endif
endfunction

" ------------------------------------------------------------------------------
" Paste options (see s:Paste() for details)

" Standard ptions
let s:stdPastePart = 1
let s:stdPasteCutout = 0
let s:stdPasteUD = 0
let s:stdPasteBF = 0

" Default options to use when drawx#cPasteOptionsReset() is called
let s:defaultPastePart = 1
let s:defaultPasteCutout = 0
let s:defaultPasteUD = 0
let s:defaultPasteBF = 0

" Paste options set by the function drawx#cPasteOptionsSet()
let s:pastePart = s:defaultPastePart
let s:pasteCutout = s:defaultPasteCutout
let s:pasteUD = s:defaultPasteUD
let s:pasteBF = s:defaultPasteBF

" ------------------------------------------------------------------------------
" Set paste options
"
" part - Indicates how the clip is to be pasted. See s:Paste() for details -
" 'tl', 'tr', 'br', bl'
" cutout - 'c' = omit leading and trailing whitespace from the pasted clip
" ud - 'u' = Paste clip upside-down
" bf - 'b' = Paste clip back-to-front
"
let s:pasteOptionsPart = ['.', 'tl', 'tr', 'br', 'bl']
let s:pasteOptionsCutout = ['no-cutout', 'cutout']
let s:pasteOptionsUD = ['right-way-up', 'upside-down']
let s:pasteOptionsBF = ['right-way-round', 'back-to-front']

" Helper function
function s:pasteOpt(new, check, old)
  let l:val = a:old
  if a:new ==# a:check
    let l:val = 1
  elseif a:new ==# '-'
    let l:val = 0
  elseif a:new !=# '.'
    let l:val = -1
  endif

  return l:val
endfunction

function drawx#cPasteOptionsSet(...)
  let l:part = 0
  let l:cutout = 0
  let l:ud = 0
  let l:bf = 0

  let l:err = 1

  if a:0
    if a:0 >= 1
      " part
      let l:part = index(s:pasteOptionsPart, a:1)
      if !l:part
        let l:part = s:pastePart
      endif
      if a:0 >= 2
        " cutout
        let l:cutout = s:pasteOpt(a:2, 'c', s:pasteCutout)
        if a:0 >= 3
          " ud
          let l:ud = s:pasteOpt(a:3, 'u', s:pasteUD)
          if a:0 >= 4
            " bf
            let l:bf = s:pasteOpt(a:4, 'b', s:pasteBF)
          endif
        endif
      endif
    endif

    let l:err = (l:part < 0) || (l:cutout < 0) || (l:ud < 0) || (l:bf < 0)

    if !l:err
      let s:pastePart = l:part
      let s:pasteCutout = l:cutout
      let s:pasteUD = l:ud
      let s:pasteBF = l:bf

      call s:PrintStatusMsg('DrawX Paste Options: '.s:pasteOptionsPart[s:pastePart].', '.s:pasteOptionsCutout[s:pasteCutout].', '.s:pasteOptionsUD[s:pasteUD].', '.s:pasteOptionsBF[s:pasteBF])
    endif
  endif

  if l:err
    call s:PrintWarningMsg("usage: DrawXPasteOpt {'tl|tr|br|bl|.' {'c|-|.' {'u|-|.' {'b|-|.'}}}} (t/b-l/r = part, c = cutout, u = upside-down, b = back-to-front : '-' = option off, '.' = don't change)")
  endif
endfunction

" Reset paste options to default
function drawx#cPasteOptionsReset(...)
  let s:pastePart = s:defaultPastePart
  let s:pasteCutout = s:defaultPasteCutout
  let s:pasteUD = s:defaultPasteUD
  let s:pasteBF = s:defaultPasteBF
  call s:PrintStatusMsg('DrawX Paste Options: '.s:pasteOptionsPart[s:pastePart].', '.s:pasteOptionsCutout[s:pasteCutout].', '.s:pasteOptionsUD[s:pasteUD].', '.s:pasteOptionsBF[s:pasteBF])
endfunction

" ------------------------------------------------------------------------------
" Paste (normal mode)
"
" std - 1 = use standard values, 0 = use values set by drawx#cPasteOptionsSet()
"
function drawx#nPaste(std)
  " Set up the position/area depending on the 'part' argument
  let l:part = (a:std) ? s:stdPastePart : s:pastePart
  if l:part == 1
    let l:posn = s:getCursorPosn()
    call extend(l:posn, [line('$'), l:posn[1] + 99999])
  elseif l:part == 2
    let l:posn = s:getCursorPosn()
    call extend(l:posn, [line('$'), l:posn[1]])
    let l:posn[1] = 1
  elseif l:part == 3
    let l:posn = [1, 1]
    call extend(l:posn, s:getCursorPosn())
  else
    let l:posn = s:getCursorPosn()
    call extend(l:posn, [l:posn[0], l:posn[1] + 99999])
    let l:posn[0] = 1
  endif

  if a:std
    call s:Paste(v:register, l:posn, l:part, s:stdPasteCutout, s:stdPasteUD, s:stdPasteBF)
  else
    call s:Paste(v:register, l:posn, l:part, s:pasteCutout, s:pasteUD, s:pasteBF)
  endif
endfunction

" ------------------------------------------------------------------------------
" Paste (visual selection)
"
" std - 1 = use standard values, 0 = use values set by drawx#cPasteOptionsSet()
"
function drawx#vPaste(std) range
  " Make a note of v:register because it will be lost as a result of callling getVOrdinates()
  let l:register = v:register
  let l:posn = s:getVOrdinates(1)
  if !empty(l:posn)
    if a:std
      call s:Paste(l:register, l:posn, s:stdPastePart, s:stdPasteCutout, s:stdPasteUD, s:stdPasteBF)
    else
      call s:Paste(l:register, l:posn, s:pastePart, s:pasteCutout, s:pasteUD, s:pasteBF)
    endif
  endif
endfunction

" ------------------------------------------------------------------------------
" Trace a path
"
" Initialise/cancel any active direction and deferred special character placement
"
function s:traceDirCancel()
  let s:traceLastDir = -1
  let s:deferredSpecial = ''
  let s:specialNoDraw = 0
endfunction

" Initialise trace operation
"
function s:tracePathInit()
  call s:traceDirCancel()
  call s:PrintWarningMsg("DrawX: Start Drawing Mode".((s:eraseMode) ? " - Erase mode is on" : ''))
endfunction

" Exit trace operation
function s:tracePathEnd()
  call s:PrintWarningMsg("DrawX: Stop Drawing Mode")
endfunction

" Convert a dashed line character to a plain one
"
" char - Character to convert
"
" Returns original character if not a dashed one, else returns the plain
" equivalent
"
function s:convertDottedChar(char)
  " If a:char is dashed then convert to plain
  let l:char = a:char
  let l:idx = index(s:styleThinDashed, a:char)
  if l:idx  >= 0
    let l:char = s:styleThin[and(l:idx, 1)]
  else
    let l:idx = index(s:styleThickDashed, a:char)
    if l:idx  >= 0
      let l:char = s:styleThick[and(l:idx, 1)]
    endif
  endif

  return l:char
endfunction

" Loopup character in specified 'join' table
"
" join - The join table to lookup in
" cchar - The character to look up
"
" Returns index of located character (or -1 if not found)
"
function s:lookupJoin(join, cchar)
  let l:idx = -1
  let l:row = 0
  while (l:idx < 0) && (l:row < len(a:join))
    let l:idx = index(a:join[l:row], a:cchar)
    let l:row += 1
  endwhile

  return l:idx
endfunction

" Interpret special character selection
"
" type - The type - 't, 'T', 'x' or 'X'
" weight - The weight of the path (0 = ASCII, 1 = thin, 2 = thick, 3 = double, 4 = dotted, 5 = blank)
" dir - The direction to consider in the selection. Some 'type' values need this
"       in order to make the correct selection. -1 = no last direction, 0 = left,
"       1 = right, 2 = up, 3 = down
" char - The current character that will be replaced by this new one
"
" Returns the selected character (or '' if unable to, owing to lack of direction)
"
function s:translateSpecialChar(type, weight, dir, char)
  let l:char = ''
  if a:type == 't'
    if a:dir >= 0
      let l:cchar = s:convertDottedChar(a:char)

      " Determine type of T character to return
      if a:dir <= 1
        " Left/right
        let l:idx = s:lookupJoin(s:joinV, l:cchar)
        let l:char = s:joinVindex[a:weight][a:dir][l:idx + 1]
      else
        " Up/down
        let l:idx = s:lookupJoin(s:joinH, l:cchar)
        let l:char = s:joinHindex[a:weight][a:dir - 2][l:idx + 1]
      endif
    endif
  elseif a:type == 'T'
    if a:dir >= 0
      " Determine type of T character to return
      if a:dir <= 1
        " Left/right
        let l:char = s:joinVindex[a:weight][a:dir][0]
      else
        " Up/down
        let l:char = s:joinHindex[a:weight][a:dir - 2][0]
      endif
    endif
  elseif a:type == 'x'
    let l:cchar = s:convertDottedChar(a:char)

    " Determine type of X character to return
    " If there is no direction, try to guess the direction that user is likely
    " to move after placing this X character - the guess is that if the current
    " character is a vertical line then the next direction will be horizontal
    " and visa-versa
    let l:idx = lookupJoin(s:joinV, l:cchar)
    if ((a:dir >= 0) && (a:dir <= 1)) || ((a:dir < 0) && (l:idx >= 0))
      " Left/right
      let l:char = s:joinVindex[a:weight][2][l:idx + 1]
    else
      " Up/down
      let l:idx = lookupJoin(s:joinH, l:cchar)
      let l:char = s:joinHindex[a:weight][2][l:idx + 1]
    endif
  elseif a:type == 'X'
    " Determine type of X character to return. The direction doesn't matter; the
    " character is the same for a particular weight
    let l:char = s:joinVindex[a:weight][2][0]
  endif

  return l:char
endfunction

" Move the cursor
"
" dir - Direction 0 = left, 1 = right, 2 = up, 3 = down
"
" Returns
" If line has been extended to accomodate new position, returns 1, else returns 0
"
function s:moveCursor(dir)
  let l:posn = s:getCursorPosn()

  if a:dir == 2
    let l:posn[0] -= 1
    if s:Drawx_autoExpandUp && !l:posn[0]
      " Have fallen off top of text - add a new line
      call append(line('.') - 1, '')
      let l:posn[0] = 1
    endif
  elseif a:dir == 1
    let l:posn[1] += 1
  elseif a:dir == 3
    let l:posn[0] += 1
    if s:Drawx_autoExpandDown && (l:posn[0] > line('$'))
      " Have fallen off bottom of text - add a new line
      call append(line('.'), '')
    endif
  else
    let l:posn[1] -= 1
  endif

  let l:extended = 0

  " Make sure we don't fall off the top/left/bottom of legal range
  if l:posn[0] && l:posn[1] && (l:posn[0] <= line('$'))
    let l:extended = s:setCursorPosnExt(l:posn[0], l:posn[1])
  endif

  return l:extended
endfunction

" Move cursor without tracing a path
"
" dir - Direction 0 = left, 1 = right, 2 = up, 3 = down
"
function s:movePath(dir)
  " Cancel the direction and deferred character placement - doesn't make sense to keep it
  call s:traceDirCancel()

  " Move the cursor
  let l:extended = s:moveCursor(a:dir)

  " This is a horrible hack! It seems that unless a change is made to the
  " buffer, the cursor position change is not indicated on the display. To get
  " around this, replace the character at the (new) cursor position with the one
  " that is already there, and then undo the change if the line was not extended
  " to accomodate the new cursor position (if we undo the change in the case
  " where the line has been extended then in some cases the line extension will
  " get removed again (I think this depends on vim/nvim version and console key
  " codes, but I'm not sure)
  let l:char = s:getCharAtCursor()
  exe 'normal! r'.l:char.((l:extended) ? '' : 'u')
endfunction

" Trace a path
"
" dir - Direction 0 = left, 1 = right, 2 = up, 3 = down
"
function s:tracePath(dir)
  if s:eraseMode
    " Erase mode is enabled

    " Cancel the direction and deferred character placement - doesn't make sense to keep it
    call s:traceDirCancel()

    " If the cursor is in virtual space then there is nothing to erase
    let l:posn = s:getCursorPosn()
    let l:virt = s:isPosnVirtual(l:posn[0], l:posn[1])
    if !l:virt
      " Erase character under cursor
      exe 'normal! r '
    endif

    " Move the cursor
    call s:moveCursor(a:dir)
  else
    " Erase mode not enabled - normal drawing mode
    let l:style = s:cs

    " Select character to use to draw trace and draw it
    let l:char = ''
    if s:specialNoDraw
      " Don't re-draw character position.
      " Note that we shouldn't actually need to draw anything but if we don't then
      " the diplay will not be updated (see comment in s:movePath()). The
      " easiest way to get around this is to re-draw the character at the cursor
      let l:char = s:getCharAtCursor()
      let s:specialNoDraw = 0
    elseif s:deferredSpecial != ''
      " There is a special character that has been deferred because of lack of a
      " direction - deal with it. We need to reverse the direction that we specify
      " to s:translateSpecialChar()
      let l:dir = [1, 0, 3, 2]

      " The cursor is guaranteed NOT to be in virtual space at this point because of previous
      " call to s:traceSpecial()
      let l:char = s:translateSpecialChar(s:deferredSpecial, l:style[4], l:dir[a:dir], s:getCharAtCursor())
      let s:deferredSpecial = ''
    else
      " Normal trace
      if a:dir == 2
        let l:char = ((s:traceLastDir == 1) ? l:style[2][2] : ((s:traceLastDir == 0) ? l:style[2][3] : l:style[1][1]))
      elseif a:dir == 1
        let l:char = ((s:traceLastDir == 2) ? l:style[2][0] : ((s:traceLastDir == 3) ? l:style[2][3] : l:style[1][0]))
      elseif a:dir == 3
        let l:char = ((s:traceLastDir == 1) ? l:style[2][1] : ((s:traceLastDir == 0) ? l:style[2][0] : l:style[1][1]))
      else
        let l:char = ((s:traceLastDir == 2) ? l:style[2][1] : ((s:traceLastDir == 3) ? l:style[2][2] : l:style[1][0]))
      endif

      " If the cursor is in virtual space the 'r'(eplace) won't work. So we have to
      " ensure it has something to work on
      let l:posn = s:getCursorPosn()
      call s:setCursorPosnExt(l:posn[0], l:posn[1])
    endif

    " Draw the selected character
    exe 'normal! r'.l:char

    " Remember the direction for next time
    let s:traceLastDir = a:dir

    " Move the cursor
    call s:moveCursor(a:dir)
  endif
endfunction

" Add special character to path (T's and X's)
"
function s:traceSpecial(type)
  let l:style = s:cs

  if (l:style[4] >= 0)
    " If the cursor is in virtual space then getCharAtCursor() will return the
    " wrong answer. Also, the 'r'(eplace) won't work. So we have to ensure it
    " has something to work on
    let l:posn = s:getCursorPosn()
    call s:setCursorPosnExt(l:posn[0], l:posn[1])

    let s:deferredSpecial = ''
    let s:specialNoDraw = 0

    " Select character to use to draw trace and draw it
    let l:char = s:translateSpecialChar(a:type, l:style[4], s:traceLastDir, s:getCharAtCursor())
    if l:char == ''
      " Cannot place character - defer
      let s:deferredSpecial = a:type
    else
      " Draw the selected character
      exe 'normal! r'.l:char

      if (s:traceLastDir >= 0)
        " We have a trace direction in force - move the cursor. If we didn't
      " do this then the next character would overwrite this one unless the user
      " breaks-out of the trace mode, moves the cursor and (if they want) re-enter
      " trace mode (basically, it's inconvenient if we don't do this!)
        call s:moveCursor(s:traceLastDir)
      else
        " No trace direction in force. Prevent further character placement at this position
        let s:specialNoDraw = 1
      endif
    endif
  else
    call s:PrintStatusMsg('Cannot place "'.toupper(a:type).'" because style set, based on "'.s:cs[0][0].'"')
  endif
endfunction

" ------------------------------------------------------------------------------
" Add special character to path (from menu)
"
function s:traceSpecialAdd(idx)
  " If the cursor is in virtual space the 'r'(eplace) won't work. So we have to
  " ensure it has something to work on
  let l:posn = s:getCursorPosn()
  call s:setCursorPosnExt(l:posn[0], l:posn[1])

  " Draw the selected character
  exe 'normal! r'.s:specialChars[a:idx]

  let s:deferredSpecial = ''
  let s:specialNoDraw = 0

  " Move the cursor
  if (s:traceLastDir >= 0)
    call s:moveCursor(s:traceLastDir)
  else
    " No trace direction in force. Prevent further character placement at this position
    let s:specialNoDraw = 1
  endif
endfunction

" ------------------------------------------------------------------------------
" Set the style from a single character
"
" char - The character to construct a style from
"
function s:setStyleFromChar(char)
  " Establish style
  let l:style = [a:char, a:char, a:char, a:char, a:char, a:char]
  let s:cs = [l:style, l:style[0:1], l:style[2:5], s:styleFills[0], -1, 0]

  " Cancel the direction and deferred character placement - doesn't make sense to keep it
  call s:traceDirCancel()
endfunction

" ------------------------------------------------------------------------------
" Construct a style based on a special character
"
function s:makeStyleFromSpecial(idx)
  call s:setStyleFromChar(s:specialChars[a:idx])
  call s:PrintStatusMsg('DrawX: Style set, based on "'.s:specialChars[a:idx].'"')
endfunction

" ------------------------------------------------------------------------------
" Erase toggle
"
function s:eraseToggle()
  let s:eraseMode = 1 - s:eraseMode
  call s:PrintStatusMsg('Erase mode: '.((s:eraseMode) ? "on" : "off"))
endfunction

" ------------------------------------------------------------------------------
" Set a style based on a supplied single character
function drawx#SetStyleFromChar(char)
  if (strchars(a:char) == 1)
    call s:setStyleFromChar(a:char)
    call s:PrintStatusMsg('DrawX: Style set, based on "'.a:char.'"')
  else
    call s:PrintWarningMsg("Usage: DrawXStyle <char>")
  endif
endfunction

" ------------------------------------------------------------------------------
" Display sub-menu selection
"
" desc - Textual description of menu
" sel - First character in the menu selection (ie, the key to press)
" list - List of characters making up the menu
"
function s:subMenuInit(desc, sel, list)
  let l:desc = a:desc.' - '
  let l:idx = 0
  for l:char in a:list
    let l:desc = l:desc.(nr2char(char2nr(a:sel) + l:idx)).":'".l:char."' "
    let l:idx += 1
  endfor

  call s:PrintStatusMsg(l:desc)
endfunction

" Sub-menu abort
"
function s:subMenuAbort()
  call s:PrintStatusMsg('Back to drawing mode')
endfunction

" ------------------------------------------------------------------------------
" Sequential command executor
"
" expr - The expression to execute (usually a function call)
" cmdKey - The command keys to inject into the comamnd buffer after the
"          expression has been evaluated. Specify '' if not required
"
function drawx#cmdExecute(expr, cmdKey)
  " Invoke operation
  if a:expr != ''
    exe a:expr
  endif
  if a:cmdKey != ''
    call feedkeys(a:cmdKey)
  endif
endfunction

" ----------------------------------
" Top-level help
"
function s:help()
  call s:PrintStatusMsg("Line weight:'".s:Drawx_lineWeightChar."' Line type:'".s:Drawx_lineTypeChar."' Corner type:'".s:Drawx_cornerChar."' Set style from special: '".s:Drawx_specialStyleChar."' Special:'".s:Drawx_specialChar."' T:'t' X:'x'")
endfunction

" ----------------------------------
" Key mapping constants
"
let g:Drawx_leaderLineWeight = g:Drawx_leaderBase.s:Drawx_lineWeightChar
let g:Drawx_leaderLineWeighta = g:Drawx_leaderBase.s:Drawx_leaderBasea.s:Drawx_lineWeightChar
let g:Drawx_leaderLineType = g:Drawx_leaderBase.s:Drawx_lineTypeChar
let g:Drawx_leaderLineTypea = g:Drawx_leaderBase.s:Drawx_leaderBasea.s:Drawx_lineTypeChar
let g:Drawx_leaderCornerType = g:Drawx_leaderBase.s:Drawx_cornerChar
let g:Drawx_leaderCornerTypea = g:Drawx_leaderBase.s:Drawx_leaderBasea.s:Drawx_cornerChar
let g:Drawx_leaderSpecial = g:Drawx_leaderBase.s:Drawx_specialChar
let g:Drawx_leaderSpeciala = g:Drawx_leaderBase.s:Drawx_leaderBasea.s:Drawx_specialChar
let g:Drawx_leaderSpecialStyle = g:Drawx_leaderBase.s:Drawx_specialStyleChar
let g:Drawx_leaderSpecialStylea = g:Drawx_leaderBase.s:Drawx_leaderBasea.s:Drawx_specialStyleChar

" A dummy 'visual yank to blackhole register' operation. This is inserted by
" some of the key mappings to force returning focus to the window after outputting
" text to the command line (this operation was chosen because it doesn't modify the
" buffer and is reliable. It DOES alter the visual selection area of course, but
" that's probably not a problem in the contexts it's used; a non-visual 'yank' does
" not work)
let s:nullCmd = 'v"_y'

" ----------------------------------
" Module Initialisation
"
function s:initialise()
  " Initialise trace direction control
  call s:traceDirCancel()

  " Set default paste options
  if exists('g:Drawx_defaultPasteOpt')
    let l:part = index(s:pasteOptionsPart, g:Drawx_defaultPasteOpt[0])
    if !l:part
      let l:part = s:defaultPastePart
    endif
    let l:cutout = s:pasteOpt(g:Drawx_defaultPasteOpt[1], 'c', s:defaultPasteCutout)
    let l:ud = s:pasteOpt(g:Drawx_defaultPasteOpt[2], 'u', s:defaultPasteUD)
    let l:bf = s:pasteOpt(g:Drawx_defaultPasteOpt[3], 'b', s:defaultPasteBF)

    if (l:part >= 0) && (l:cutout >= 0) && (l:ud >= 0) && (l:bf >= 0)
      let s:defaultPastePart = l:part
      let s:defaultPasteCutout = l:cutout
      let s:defaultPasteUD = l:ud
      let s:defaultPasteBF = l:bf

      let s:pastePart = s:defaultPastePart
      let s:pasteCutout = s:defaultPasteCutout
      let s:pasteUD = s:defaultPasteUD
      let s:pasteBF = s:defaultPasteBF
    else
      call s:PrintWarningMsg('DrawX: Error interpreting g:Drawx_defaultPasteOpt')
    endif
  endif

  " Key mappings for cursor movement
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."H".' :call drawx#cmdExecute("call s:movePath(0)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."L".' :call drawx#cmdExecute("call s:movePath(1)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."K".' :call drawx#cmdExecute("call s:movePath(2)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."J".' :call drawx#cmdExecute("call s:movePath(3)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."<C-left>".' :call drawx#cmdExecute("call s:movePath(0)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."<C-right>".' :call drawx#cmdExecute("call s:movePath(1)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."<C-up>".' :call drawx#cmdExecute("call s:movePath(2)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."<C-down>".' :call drawx#cmdExecute("call s:movePath(3)", '."'".g:Drawx_leaderBase."'".')<cr>'

  " Key mappings for cursor movement with drawing
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."h".' :call drawx#cmdExecute("call s:tracePath(0)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."l".' :call drawx#cmdExecute("call s:tracePath(1)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."k".' :call drawx#cmdExecute("call s:tracePath(2)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."j".' :call drawx#cmdExecute("call s:tracePath(3)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."<left>".' :call drawx#cmdExecute("call s:tracePath(0)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."<right>".' :call drawx#cmdExecute("call s:tracePath(1)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."<up>".' :call drawx#cmdExecute("call s:tracePath(2)", '."'".g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase."<down>".' :call drawx#cmdExecute("call s:tracePath(3)", '."'".g:Drawx_leaderBase."'".')<cr>'

  " Key mappings for line weight selection
  exe 'nnoremap <silent> '.g:Drawx_leaderLineWeight.' :call drawx#cmdExecute("call s:subMenuInit('."'Line weight', '0'".', s:lineWeightChars)", '."'".s:nullCmd.g:Drawx_leaderLineWeighta."'".')<cr>'
  let l:idx = 0
  for l:char in s:lineWeightChars
    exe 'nnoremap <silent> '.g:Drawx_leaderLineWeighta.(nr2char(char2nr('0') + l:idx)).' :call drawx#cmdExecute("call s:setStyle('.l:idx.', -1, -1)", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'
    let l:idx += 1
  endfor
  exe 'nnoremap <silent> '.g:Drawx_leaderLineWeighta.s:Drawx_exitMenu.' :call drawx#cmdExecute("call s:subMenuAbort()", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'

  " Key mappings for line type selection
  exe 'nnoremap <silent> '.g:Drawx_leaderLineType.' :call drawx#cmdExecute("call s:subMenuInit('."'Line type', '1'".', s:lineTypeChars)", '."'".s:nullCmd.g:Drawx_leaderLineTypea."'".')<cr>'
  let l:idx = 0
  for l:char in s:lineTypeChars
    exe 'nnoremap <silent> '.g:Drawx_leaderLineTypea.(nr2char(char2nr('1') + l:idx)).' :call drawx#cmdExecute("call s:setStyle(-1, '.(1 + l:idx).', -1)", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'
    let l:idx += 1
  endfor
  exe 'nnoremap <silent> '.g:Drawx_leaderLineTypea.s:Drawx_exitMenu.' :call drawx#cmdExecute("call s:subMenuAbort()", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'

  " Key mappings for corner type selection
  exe 'nnoremap <silent> '.g:Drawx_leaderCornerType.' :call drawx#cmdExecute("call s:subMenuInit('."'Corner type', 'a'".', s:cornerTypeChars)", '."'".s:nullCmd.g:Drawx_leaderCornerTypea."'".')<cr>'
  let l:idx = 0
  for l:char in s:cornerTypeChars
    exe 'nnoremap <silent> '.g:Drawx_leaderCornerTypea.(nr2char(char2nr('a') + l:idx)).' :call drawx#cmdExecute("call s:setStyle(-1, -1, '.(1 + l:idx).')", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'
    let l:idx += 1
  endfor
  exe 'nnoremap <silent> '.g:Drawx_leaderCornerTypea.s:Drawx_exitMenu.' :call drawx#cmdExecute("call s:subMenuAbort()", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'

  " Key mappings for special characters
  exe 'nnoremap <silent> '.g:Drawx_leaderSpecial.' :call drawx#cmdExecute("call s:subMenuInit('."'Special', 'a'".', s:specialChars)", '."'".s:nullCmd.g:Drawx_leaderSpeciala."'".')<cr>'
  let l:idx = 0
  for l:char in s:specialChars
    exe 'nnoremap <silent> '.g:Drawx_leaderSpeciala.(nr2char(char2nr('a') + l:idx)).' :call drawx#cmdExecute("call s:traceSpecialAdd('.l:idx.')", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'
    let l:idx += 1
  endfor
  exe 'nnoremap <silent> '.g:Drawx_leaderSpeciala.s:Drawx_exitMenu.' :call drawx#cmdExecute("call s:subMenuAbort()", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'

  " Key mappings for setting style based on a special character
  exe 'nnoremap <silent> '.g:Drawx_leaderSpecialStyle.' :call drawx#cmdExecute("call s:subMenuInit('."'Set style from special', 'a'".', s:specialChars)", '."'".s:nullCmd.g:Drawx_leaderSpecialStylea."'".')<cr>'
  let l:idx = 0
  for l:char in s:specialChars
    exe 'nnoremap <silent> '.g:Drawx_leaderSpecialStylea.(nr2char(char2nr('a') + l:idx)).' :call drawx#cmdExecute("call s:makeStyleFromSpecial('.l:idx.')", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'
    let l:idx += 1
  endfor
  exe 'nnoremap <silent> '.g:Drawx_leaderSpecialStylea.s:Drawx_exitMenu.' :call drawx#cmdExecute("call s:subMenuAbort()", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'

  " Key mappings for t/T and x/X
  exe 'nnoremap <silent> '.g:Drawx_leaderBase.'t :call drawx#cmdExecute("call s:traceSpecial('."'t'".')", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase.'x :call drawx#cmdExecute("call s:traceSpecial('."'x'".')", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase.'T :call drawx#cmdExecute("call s:traceSpecial('."'T'".')", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'
  exe 'nnoremap <silent> '.g:Drawx_leaderBase.'X :call drawx#cmdExecute("call s:traceSpecial('."'X'".')", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'

  " Key mapping for direction cancel
  exe 'nnoremap <silent> '.g:Drawx_leaderBase.'! :call drawx#cmdExecute("call s:traceDirCancel()", '."'".g:Drawx_leaderBase."'".')<cr>'

  " Key mapping for erase toggle
  exe 'nnoremap <silent> '.g:Drawx_leaderBase.'<space> :call drawx#cmdExecute("call s:eraseToggle()", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'

  " Help
  exe 'nnoremap <silent> '.g:Drawx_leaderBase.'? :call drawx#cmdExecute("call s:help()", '."'".s:nullCmd.g:Drawx_leaderBase."'".')<cr>'

  " Escape/exit trace mode
  exe 'nnoremap <silent> '.g:Drawx_leaderBase.s:Drawx_exitTrace.' :call <sid>tracePathEnd()<cr>'
endfunction

call s:initialise()

" ------------------------------------------------------------------------------
" eof

